#include <avr/io.h>
#ifndef F_CPU
#define F_CPU 16000000
#endif

#include <util/delay.h>
#include <stdlib.h>
#include <stdbool.h>
#include <inttypes.h>
#include "./usart/usart.h"
#include "./spi/spi.h"
#include "./ST7735R/ST7735R_TFT.h"

#define LCD_MAX_X ST7735R_WIDTH
#define LCD_MAX_Y ST7735R_HEIGHT
#define SNAKE_SIZE 120
#define SNAKE_MAX_DIM 10

#define CELL_WITDH 5
#define CELL_HEIGTH 5
#define FOOD_SIZE 				 6
#define SPACE_SNAKES			 5
#define HEADER_OFFSET 			 10

typedef struct cell {
	int x;
	int y;
	int size;
} cell;

typedef struct {
	int x;
	int y;
} Point;

typedef struct {
	cell cells[SNAKE_SIZE];
	int curr_dim;
} Snake;

enum Food_status {
	EATEN,
	RECENTLY_EATEN,
	NOT_EATEN
};

enum Menu_option {
	NEW_GAME,
	LEVEL,
	EXIT
};

enum Game_mode {
	BOX = 1,
	NOBOX
};

typedef struct {
	int x;
	int y;
	enum Food_status food_status;
} Food;

enum Game_status {
	STATE_MENU = 1,
	STATE_GAME_RUNNING,
	STATE_GAME_ENDED
};

enum Game_result {
	GAME_LOST = 1,
	GAME_WON
};

enum direction_input {
	DOWN = 2,
	UP = 1,
	LEFT = 4,
	RIGHT = 3
};

static Snake snake;
static Food food;
static enum Game_status game_status;
static enum direction_input direction;
static enum Game_result game_result;
static enum Menu_option menu_option;
static int SNAKE_SPEED = 2;
static enum Game_mode game_mode;

/* generate food */
void food_generation() {
	if(food.food_status == NOT_EATEN || food.food_status == RECENTLY_EATEN)
		return;

	food.food_status = NOT_EATEN;
	int good_pos = 0;
	int x;
	int y;
	do {
		good_pos = 1;
		x = (rand() % (LCD_MAX_X - 10)) + 5;
		y = (rand() % (LCD_MAX_Y - 10)) + HEADER_OFFSET;

		if(food.x == x && food.y == y)
			good_pos = 0;

		for(int i = 0; i < snake.curr_dim; i++) {
			if(snake.cells[i].x == x && snake.cells[i].y == y) {
				good_pos = 0;
			}
		}
	} while(good_pos == 0);

	food.x = x;
	food.y = y;
}

static void draw_header(void) {
	char show_score[50];
	char level[50];

	sprintf(show_score, "Score: %d", snake.curr_dim - 2);
	ST7735R_DrawText(5, 6, show_score, 255, 255, 255, 0, 0, 0);

	sprintf(level, "Level: %d", SNAKE_SPEED - 1);
	ST7735R_DrawText(65, 6, level, 255, 255, 255, 0, 0, 0);

	ST7735R_Line(0, 16, LCD_MAX_X, 16, 255, 255, 255);
}

int valid_direction(enum direction_input dir) {
	if(snake.curr_dim == 1) {
		return 1;
	}

	if(dir == UP && snake.cells[1].y == snake.cells[0].y + SNAKE_SPEED)
		return 0;
	else if(dir == DOWN && snake.cells[1].y == snake.cells[0].y - SNAKE_SPEED)
		return 0;
	else if(dir == LEFT && snake.cells[1].x == snake.cells[0].x - SNAKE_SPEED)
		return 0;
	else if(dir == RIGHT && snake.cells[1].x == snake.cells[0].x + SNAKE_SPEED)
		return 0;

	return 1;
}

/* update the snake coordoantes */
void update_snake_position(enum direction_input dir) {
	cell new_head;
	new_head.x = snake.cells[0].x;
	new_head.y = snake.cells[0].y;

	switch(dir) {
		case UP:
			new_head.y = new_head.y + SNAKE_SPEED + SPACE_SNAKES;
			break;
		case DOWN:
			new_head.y = new_head.y - SNAKE_SPEED - SPACE_SNAKES;
			break;
		case LEFT:
			new_head.x = new_head.x - SNAKE_SPEED - SPACE_SNAKES;
			break;
		case RIGHT:
			new_head.x = new_head.x + SNAKE_SPEED + SPACE_SNAKES;
			break;
	}

	if (new_head.x >= LCD_MAX_X) {
		new_head.x = 3 + SPACE_SNAKES;
	} else if (new_head.x <= 0) {
		new_head.x = LCD_MAX_X - 3 - SPACE_SNAKES;
	} else if (new_head.y >= LCD_MAX_Y) {
		new_head.y = HEADER_OFFSET + SPACE_SNAKES;
	} else if (new_head.y <= HEADER_OFFSET) {
		new_head.y = LCD_MAX_Y - 3 - SPACE_SNAKES;
	}

	for(int i = snake.curr_dim - 1; i > 0; i--) {
		snake.cells[i].x = snake.cells[i - 1].x;
		snake.cells[i].y = snake.cells[i - 1].y;
	}

	snake.cells[0].x = new_head.x;
	snake.cells[0].y = new_head.y;
}

static void draw_cell(int x, int y, uint8_t r, uint8_t g, uint8_t b)
{
	ST7735R_FillRect(x, y, x +
		CELL_WITDH, y + CELL_HEIGTH, r, g, b);
}

static void draw_snake()
{
	for (int i = 0; i < snake.curr_dim; i++) {
		draw_cell(snake.cells[i].x, snake.cells[i].y,
			0, 255, 0);
	}
}

static void draw_food()
{
	if (food.food_status == RECENTLY_EATEN) {
		ST7735R_FillRect(food.x, food.y, food.x +
			FOOD_SIZE, food.y + FOOD_SIZE, 0, 0, 0);
		food.food_status = EATEN;
	} else if(food.food_status == NOT_EATEN) {
		ST7735R_FillRect(food.x, food.y, food.x +
			FOOD_SIZE, food.y + FOOD_SIZE, 255, 0, 0);
	}
}

void draw_menu() {
	/* draw title */
	char title[] = "SNAKE GAME";
	ST7735R_DrawText(40, 20, title, 255, 255, 255, 0, 0, 0);

	/* draw new game */
	char new_game[] = "NEW GAME";
	ST7735R_DrawText(40, 60, new_game, 255, 255, 255, 0, 0, 0);

	char level[50];
	sprintf(level, "LEVEL: %d", SNAKE_SPEED - 1);
	ST7735R_DrawText(40, 80, level, 255, 255, 255, 0, 0, 0);

	// char mod[50];
	// sprintf(mod, "LEVEL: %s", SNAKE_SPEED - 1);
	// ST7735R_DrawText(40, 80, level, 255, 255, 255, 0, 0, 0);

	char ex[] = "EXIT";
	ST7735R_DrawText(40, 100, ex, 255, 255, 255, 0, 0, 0);

	switch(menu_option) {
		case NEW_GAME:
			ST7735R_FilledCircle(30, 60, 1, 255, 0, 0);
			ST7735R_FilledCircle(30, 80, 1, 0, 0, 0);
			ST7735R_FilledCircle(30, 100, 1, 0, 0, 0);
			break;
		case LEVEL:
			ST7735R_FilledCircle(30, 80, 1, 255, 0, 0);
			ST7735R_FilledCircle(30, 60, 1, 0, 0, 0);
			ST7735R_FilledCircle(30, 100, 1, 0, 0, 0);
			break;
		case EXIT:
			ST7735R_FilledCircle(30, 100, 1, 255, 0, 0);
			ST7735R_FilledCircle(30, 80, 1, 0, 0, 0);
			ST7735R_FilledCircle(30, 60, 1, 0, 0, 0);
			break;
	}
}

/* implements the menu logic */
void menu_logic() {
	draw_menu();
	char received_char = blocking_USART0_receive();
	if(received_char == '1') {
		if(menu_option == EXIT)
			menu_option = LEVEL;
		else
			menu_option = NEW_GAME;
	} else if(received_char == '2') {
		if(menu_option == NEW_GAME)
			menu_option = LEVEL;
		else
			menu_option = EXIT;
	} else if(received_char == '4') {
		if(menu_option == NEW_GAME) {
			game_status = STATE_GAME_RUNNING;
			ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1,
				 ST7735R_HEIGHT - 1, 0, 0, 0);
		} else if(menu_option == LEVEL) {
			SNAKE_SPEED++;
		}
	} else if(received_char == '3' && menu_option == LEVEL && SNAKE_SPEED > 2) {
		SNAKE_SPEED--;
	}
}

/* converts char to a specific enum */
void convert_to_enum(char choice) {
	if(choice == '2' && direction != DOWN)
		direction = UP;
	else if(choice == '1' && direction != UP)
		direction = DOWN;
	else if(choice == '4' && direction != LEFT)
		direction = RIGHT;
	else if(choice == '3' && direction != RIGHT)
		direction = LEFT;
}

void buzz() {
	PORTC |= (1 << PC0);
	_delay_ms(1);
	PORTC &= ~(1 << PC0);
}

static void check_for_collison() {
	/* collision with food */
	cell snake_head;
	snake_head.x = snake.cells[0].x;
	snake_head.y = snake.cells[0].y;

	if (abs(food.x - snake_head.x) < CELL_WITDH * 2
		&& abs(food.y - snake_head.y) < CELL_HEIGTH * 2 && food.food_status != EATEN
		&& food.food_status != RECENTLY_EATEN) {
		ST7735R_FillRect(food.x, food.y, food.x +
			FOOD_SIZE, food.y + FOOD_SIZE, 0, 0, 0);
		food.food_status = RECENTLY_EATEN;
		snake.curr_dim++;
		buzz();
		food_generation();
	}

	/* check for collision with body */
	for(int i = 1; i < snake.curr_dim; i++) {
		if (abs(snake.cells[i].x - snake_head.x) * 2 < CELL_WITDH &&
			abs(snake.cells[i].y - snake_head.y) * 2 < CELL_HEIGTH) {
			game_result = GAME_LOST;
			game_status = STATE_GAME_ENDED;
			ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1,
				ST7735R_HEIGHT - 1, 0, 0, 0);
			return;
		}
	}

	if(snake.curr_dim / (SNAKE_SPEED) > 7) {
		ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1,
			ST7735R_HEIGHT - 1, 0, 0, 0);
		game_status = STATE_GAME_ENDED;
		game_result = GAME_WON;
	}
}

/* implements the logic of the game */
void game_logic() {
	char received_char = nonblocking_USART0_receive();
	if(received_char != '\0') {
		convert_to_enum(received_char);
		USART0_transmit(received_char);
	}

	draw_cell(snake.cells[snake.curr_dim - 1].x,
		snake.cells[snake.curr_dim - 1].y, 0, 0, 0);

	food_generation();
	update_snake_position(direction);
	check_for_collison();
	if(game_status == STATE_GAME_ENDED)
		return;

	draw_header();
	draw_snake();
	draw_food();

	_delay_ms(250);
}

/* implements end game logic */
void end_game_logic() {
	if(game_result == GAME_WON) {
		char result[50];
		ST7735R_DrawText(25, LCD_MAX_Y / 2 + 10,
			"YOU WON!", 255, 255, 255, 0, 0, 0);

		sprintf(result, "Final score: %d", snake.curr_dim);
		ST7735R_DrawText(25, LCD_MAX_Y / 2 - 10,
			result, 255, 255, 255, 0, 0, 0);

		ST7735R_DrawText(25, LCD_MAX_Y / 2 - 30,
			"Press ^ for menu", 255, 255, 255, 0, 0, 0);

		char received_char = blocking_USART0_receive();
		if(received_char == '1') {
			game_status = STATE_MENU;
			snake.curr_dim = 2;
			snake.cells[0].x = LCD_MAX_X / 2;
			snake.cells[0].y = LCD_MAX_Y / 2;
			direction = DOWN;
			ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1,
				ST7735R_HEIGHT - 1, 0, 0, 0);
		}
	} else {
		char result[50];
		ST7735R_DrawText(25, LCD_MAX_Y / 2 + 10,
			"YOU LOST!", 255, 255, 255, 0, 0, 0);

		sprintf(result, "Final score: %d", snake.curr_dim);
		ST7735R_DrawText(25, LCD_MAX_Y / 2 - 10,
			result, 255, 255, 255, 0, 0, 0);

		ST7735R_DrawText(25, LCD_MAX_Y / 2 - 30,
			"Press ^ for menu", 255, 255, 255, 0, 0, 0);

		char received_char = blocking_USART0_receive();
		if(received_char == '1') {
			game_status = STATE_MENU;
			snake.curr_dim = 2;
			snake.cells[0].x = LCD_MAX_X / 2;
			snake.cells[0].y = LCD_MAX_Y / 2;
			direction = DOWN;
			ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1,
				ST7735R_HEIGHT - 1, 0, 0, 0);
		}
	}
}

/* initialize all devies used */
void init_devices() {

	/* initialize the food variable */
	food.food_status = NOT_EATEN;
	food.x = LCD_MAX_X / 2;
	food.y = 20;

	game_mode = NOBOX;

	/* initialize the random generator */
	srand(1000);

	/* initialize direction when starting the game */
	direction = DOWN;

	/* init the snake structure */
	snake.curr_dim = 2;

	/* initialize the first cell of the snake */
	snake.cells[0].x = LCD_MAX_X / 2;
	snake.cells[0].y = LCD_MAX_Y / 2;

	/* initial state of the game */
	game_status = STATE_MENU;
	game_result = GAME_LOST;

	/* Setăm pinul 0 al portului C ca pin de ieșire. */
	DDRC |= (1 << PC0);
	PORTC &= ~(1 << PC0);
	/* blueetooth communication initialization */
	USART0_init();

	/* Init SPI */
	SPI_init();

	/* Init LCD */
	ST7735R_Begin();

	/* Clear screen with BLACK. */
	ST7735R_FillRect(0, 0, ST7735R_WIDTH - 1, ST7735R_HEIGHT - 1, 0, 0, 0);
}

int main() {
	init_devices();
	while(1) {
		switch(game_status) {
			case STATE_MENU:
				menu_logic();
				break;
			case STATE_GAME_RUNNING:
				game_logic();
				break;
			case STATE_GAME_ENDED:
				end_game_logic();
				break;
			default:
				menu_logic();
				break;
		}
	}
	return 0;
}
